" Name: netclient
" Description: Vim global plugin for browsing the internet
" Last Change: 2023 Jul 4
" Maintainer: Sloum <sloum@rawtext.club> 
" License: Vim license

if exists("g:loaded_netclient_global")
  finish
endif
let g:loaded_netclient_global = 1

let save_cpo = &cpo
set cpo&vim

if !exists("g:netclient_command")
  let g:netclient_command = "lynx -dump -dont_wrap_pre %s"
  " XXX
  " A client backend written in go that supports gopher,
  " gemini, and http/https called `ncg` is specifically
  " written for this plugin and is available from:
  " https://git.rawtext.club/sloum/ncg
  " If you get it, comment the above lynx line and
  " uncomment the below line to use it.
  "
  " let g:netclient_command = "ncg %s"
endif

if !exists("g:netclient_buffer_style")
  " 0 - New buffer in same window
  " 1 - New buffer in new split
  " 2 - New buffer in new tab
  let g:netclient_buffer_style = 0
endif

" Sets the local leader to '  ' (space space)
" for use in mappings within the plugin
if !exists("g:netclient_leader")
  let g:netclient_leader = "  "
endif

if !exists("g:netclient_web_search")
  let g:netclient_web_search = "https://wiby.me/?q=%s"
endif

if !exists("g:netclient_gemini_search")
  let g:netclient_gemini_search = "gemini://geminispace.info:1965/search?%s"
endif

if !exists("g:netclient_gopher_search")
  let g:netclient_gopher_search = "gopher://gopher.floodgap.com:70/1/v2/vs\t%s"
endif

let maplocalleader = g:netclient_leader

function NetclientClearBuffer()
  call deletebufline(bufname(), 1, line("$"))
endfunction

function NetclientFindURL(num)
  let l:url = ""
  if NetclientIsUrlSimple(a:num)
    return a:num
  elseif a:num
    " The later part of this should verify that there is at
    " least a :// in the link, rather than just being text.
    " Otherwise there is a risk of it getting an ordered list
    " item instead of the link. In theory, that is still possible
    " if an ordered list item has nothing but a URL in it...
    " Think on this
    let l:exp = "^\\s\\+". a:num . "\\.\\s*\\(\\S\\+\\)"
    let l:line = getline(search(l:exp, "cnwb"))
    if l:line == ""
      echo "\rInvalid link number: " . a:num
      return
    endif
    let l:url = substitute(l:line, l:exp, '\=submatch(1)', '')
  else
    let l:url = expand("<cWORD>")
  endif
  return l:url
endfunction

function NetclientIsUrlSimple(u)
  let l:isURL = match(a:u, '\w\+:\/\/\S\+')
  if l:isURL >= 0
    return 1
  else
    return 0
  endif
endfunction


function NetclientFollowURLnew(num)
  let l:url = NetclientFindURL(a:num)
  if NetclientIsUrlSimple(l:url)
    if g:netclient_buffer_style
      vnew
    else
      enew
    endif
    set filetype=netclient
		call NetclientGetU(l:url)
  else
    echo "\rInvalid URL"
  endif
endfunction

function NetclientFollowURL(num)
  let l:valid_buf = NetclientIsNCbuffer()
  if !l:valid_buf
    call NetclientFollowURLnew(a:num)
    return
  endif
  let l:url = NetclientFindURL(a:num)

  if NetclientIsUrlSimple(l:url)
		call NetclientGetU(l:url)
  else
    echo "\rInvalid URL"
  endif
endfunction

function NetclientIsNCbuffer()
  let l:f = &filetype
  if l:f == "netclient"
    return 1
  elseif l:f == ""
    set filetype=netclient
    return 1
  else
    return 0
  endif
endfunction

function NetclientGoNew()
  if g:netclient_buffer_style == 1
    vnew
  elseif g:netclient_buffer_style == 2
    tabnew
  else
    enew
  endif
  set filetype=netclient
  call NetclientGo()
endfunction

function NetclientGo()
  let l:valid = NetclientIsNCbuffer()
  if !l:valid
    call NetclientGoNew()
    return
  endif
  let l:u = trim(input("GO: "))
  if l:u == ""
    return
  endif
  call NetclientGetU(l:u)
endfunction

function NetclientGetU(u)
	let l:addr = printf(g:netclient_command, shellescape(a:u))
  let l:txt = system(l:addr)
  let l:resp = substitute(l:txt, "[\\x0]", "\\n", "g")
  let l:resp_split = split(l:resp, "\n")
  let l:u = a:u
  if len(l:resp_split) > 1 && match(l:resp_split[0], '\w\+:\/\/') == 0
    let l:u = trim(l:resp_split[0])
    let l:resp_split = l:resp_split[1:]
  endif
  if v:shell_error == 0
    call NetclientClearBuffer()
    call append(0,[l:u, ""] +  l:resp_split)
    call cursor(1,1)
  elseif v:shell_error == 22
    " gemini line request
    let l:query = trim(input(l:resp))
    let l:qloc = match(l:u, "?")
    if l:qloc < 0
      call NetclientGetU(l:u . "?" . UrlEncode(l:query))
    else
      call NetclientGetU(l:u[: l:qloc] . UrlEncode(l:query))
    endif
  elseif v:shell_error == 33
    let l:query = trim(input(l:resp))
    let l:qloc = match(l:u, "\t")
    if l:qloc < 0
      call NetclientGetU(l:u . "\t" . l:query)
    else
      call NetclientGetU(l:u[: l:qloc] . l:query)
    endif
  else
    echo "\r" . l:resp
  endif
endfunction

function NetclientCheckURL(num)
  let l:exp = "^\\s\\+". a:num . "\\.\\s*\\(\\S\\+\\)"
  let l:line = getline(search(l:exp, "cnwb"))
  if l:line == ""
    echo "\rInvalid link number: " . a:num
    return
  endif
  let l:url = substitute(l:line, l:exp, '\=submatch(1)', '')
  echo "\r" . l:url
endfunction

" URL encode a string. ie. Percent-encode characters as necessary.
" From: http://www.danielbigham.ca/cgi-bin/document.pl?mode=Display&DocumentID=1053
function! UrlEncode(string)
	let result = ""
	let characters = split(a:string, '.\zs')
	for character in characters
		if character == " "
			let result = result . "%20"
		elseif CharacterRequiresUrlEncoding(character)
			let i = 0
			while i < strlen(character)
				let byte = strpart(character, i, 1)
				let decimal = char2nr(byte)
				let result = result . "%" . printf("%02x", decimal)
				let i += 1
			endwhile
		else
			let result = result . character
		endif
	endfor
	return result
endfunction

" Returns 1 if the given character should be percent-encoded in a URL encoded string.
" From: http://www.danielbigham.ca/cgi-bin/document.pl?mode=Display&DocumentID=1053
function! CharacterRequiresUrlEncoding(character)
	let ascii_code = char2nr(a:character)
	if ascii_code >= 48 && ascii_code <= 57
		return 0
	elseif ascii_code >= 65 && ascii_code <= 90
		return 0
	elseif ascii_code >= 97 && ascii_code <= 122
		return 0
	elseif a:character == "-" || a:character == "_" || a:character == "." || a:character == "~"
		return 0
	endif
	return 1
endfunction

function NetclientSearch(scheme, new)
  let l:search = trim(input("Search: "))
  if a:scheme == "http"
    let l:search = UrlEncode(trim(l:search))
    let l:search = printf(g:netclient_web_search, l:search)
    if a:new
      call NetclientFollowURLnew(l:search)
    else
      call NetclientFollowURL(l:search)
    endif
  elseif a:scheme == "gemini"
    let l:u = UrlEncode(trim(l:search))
    if a:new
      call NetclientFollowURLnew(printf(g:netclient_gemini_search, l:search))
    else
      call NetclientFollowURL(printf(g:netclient_gemini_search, l:search))
    endif
  elseif a:scheme == "gopher"
    if a:new
      call NetclientFollowURLnew(printf(g:netclient_gopher_search, l:search))
    else
      call NetclientFollowURL(printf(g:netclient_gopher_search, l:search))
    endif
  else
    echo "\rUnknown search scheme " . a:scheme
  endif
endfunction



nnoremap <localleader>sP :call NetclientSearch("gopher", 1)<cr>
nnoremap <localleader>sp :call NetclientSearch("gopher", 0)<cr>
nnoremap <localleader>sG :call NetclientSearch("gemini", 1)<cr>
nnoremap <localleader>sg :call NetclientSearch("gemini", 0)<cr>
nnoremap <localleader>sW :call NetclientSearch("http", 1)<cr>
nnoremap <localleader>sw :call NetclientSearch("http", 0)<cr>
nnoremap <localleader>O :call NetclientGoNew()<cr>
nnoremap <localleader>o :call NetclientGo()<cr>
nnoremap <localleader>f :<C-U>call NetclientFollowURL(v:count)<cr>
nnoremap <localleader>F :<C-U>call NetclientFollowURLnew(v:count)<cr>
nnoremap <localleader>u :<C-U>call NetclientCheckURL(v:count)<cr>

let &cpo = save_cpo

